require 'openssl'
require 'base64'
require 'securerandom'
# Api Authentication Module
# =========================
# <b>How it works</b>
# ============
# <i>Client Side</i>
# ===========
# A string is first created using your HTTP payload containing all parameters and a preshared key.
# This string is then used to create the signature which is a Base64 encoded SHA1 HMAC, using the clients private secret key.
# This signature is then can be added to Authorization HTTP header 
# <code>Authorization = Authy.generate_signed_header(params)</code>
# You can pass explicite service name and shared_key as an argument to support multiple API authentication
# <code>Authorization = Authy.generate_signed_header(params, 'client_1', ENV['client_1_shared_key'])</code>
# Now add this to to your header
# <code>header["Authorization"] = Authorization</code>
#
# secret_key generated from the server side can be kept inside the a environment variable with name <i>shared_key</i> along with <i>service_name</i> as your service name    
#  
# <i>Server Side API validation</i>
# <code>is_valid = Authy.validate_api(params, request.headers)</code>
# Pass explicite shared key to support multiple API authentication
# <code>is_valid = Authy.validate_api(params, request.headers,ENV['client_1_shared_key'])</code>
# Generating secret key for clients
# <code>secret_key = Authy.generate_secret_key</code>  
module Authy
  # Create a signed hash using input parameter
  #
  # Example:
  #   >> Authy.generate_signed_header({ 'key' => value })
  #   => {:"X-OWNING-SERVICE"=>"l", :"X-SIGNING-ORDER"=>"", :"X-SIGNATURE"=>""}
  #
  # Arguments:
  #   data: (map)
  def self.generate_signed_header(data, service_name = ENV['service_name'], shared_key = ENV['shared_key'])
    processed_data, deep_lookup_table = Authy.fully_qualified_keys(data)
    signing_order = processed_data + ',' + 'owning_service,signing_order'
    request_data = Authy.build_query(processed_data, deep_lookup_table)
    unsigned_data = request_data + \
                    '&owning_service=' + service_name + '&signing_order=' + \
                    signing_order
    digest = OpenSSL::Digest.new('sha1')
    hmac = OpenSSL::HMAC.digest(digest, shared_key, unsigned_data)
    signed_str = Base64.strict_encode64(hmac)
    { 'X-OWNING-SERVICE': service_name, 'X-SIGNING-ORDER': signing_order,
      'X-SIGNATURE': signed_str }
  end

  # Method to create signed order and signed parameter deep lookup map
  def self.fully_qualified_keys(object, fully_qualified_keys = [], val = '', prefix = nil, map = {}) # :nodoc:
    if object.is_a? Hash
      object.each do |key, val|
        key_ = prefix.nil? ? key : prefix + '[' + key + ']'
        Authy.fully_qualified_keys(val, fully_qualified_keys,
                                     val, key_, map)
      end
    elsif object.is_a? Array
      object.each_with_index do |item, index|
        key_ = prefix + '[' + index.to_s + ']'
        Authy.fully_qualified_keys(item, fully_qualified_keys,
                                     item, key_, map)
      end
    else
      map[prefix] = val.to_s
      fully_qualified_keys << prefix
    end
    [fully_qualified_keys.join(','), map]
  end
  def self.fully_qualified_keys_2(object, fully_qualified_keys = [], val = '', prefix = nil, map = {})
    if object.is_a? Hash
      object.each do |key, val|
        key_ = prefix.nil? ? key : prefix + '[' + key + ']'
        Authy.fully_qualified_keys_2(val, fully_qualified_keys, val, key_, map)
      end
    elsif object.is_a? Array
      object.each_with_index do |item, index|
        key_ = prefix + '[' + index.to_s + ']'
        Authy.fully_qualified_keys_2(item, fully_qualified_keys, item, key_, map)
      end
    else
      map[prefix] = val.to_s
      fully_qualified_keys << prefix
    end
    [fully_qualified_keys.join(','), map]
  end

  # Utility method to build query from deep lookup table and parameters
  def self.build_query(processed_data, deep_lookup_table) # :nodoc:
    query = []
    processed_data.split(',').each { |data| query << data + '=' + deep_lookup_table[data].to_s }
    query.join('&')
  end

  # Validate API
  #
  # Example:
  #   >> Authy.validate_api(params, request.headers)
  #   => true/false
  #
  # Arguments:
  #   params: (map)
  #   header
  def self.validate_api(params, header, shared_key = ENV['shared_key'])
    owning_service = header['Authorization'][:'X-OWNING-SERVICE']
    expected_header = Authy.generate_signed_header(params, owning_service, shared_key)
    expected_header[:'X-SIGNATURE'] == header['Authorization'][:'X-SIGNATURE']
  end

  # Generate a pre-shared key for client
  #
  # Example:
  #   >> Authy.generate_secret_key
  #   => "63SnNesbvA9IuwXlkpF9KA=="
  def self.generate_secret_key
    SecureRandom.base64
  end
end
